from typing import Union

import flask
import pathlib
from keg.web import BaseView

import keg_storage


public_bp = flask.Blueprint('public', __name__)


def create_local_storage(root: Union[str, pathlib.Path]) -> keg_storage.LocalFSStorage:
    return keg_storage.LocalFSStorage(
        root=root,
        linked_endpoint='public.link-view',
        secret_key=b'a' * 32,
        name='local.storage'
    )


class LinkView(BaseView, keg_storage.LinkViewMixin):
    blueprint = public_bp
    url = '/storage'

    def get_storage_backend(self) -> keg_storage.InternalLinksStorageBackend:
        root = flask.request.headers['StorageRoot']
        return create_local_storage(root)



import collections
import logging
import warnings
from typing import Mapping, Optional

import flask
import itsdangerous

from keg_storage import cli
from keg_storage import backends


log = logging.getLogger(__name__)


class Storage:
    """A proxy and management object for storage backends."""

    _interfaces: Mapping[str, backends.StorageBackend]
    interface: Optional[str]

    def __init__(self, app: Optional[flask.Flask] = None, cli_group_name='storage'):
        self.cli_group_name = cli_group_name
        self._interfaces = {}

        if app:
            self.init_app(app)

    def _migrate_config(self, app: flask.Flask):
        """Handle changes in configuration variable names."""

        if "STORAGE_PROFILES" in app.config:
            if "KEG_STORAGE_PROFILES" in app.config:
                warnings.warn(
                    "Found both KEG_STORAGE_PROFILES and deprecated STORAGE_PROFILES. "
                    "Please remove the obsolete configuration.",
                    DeprecationWarning,
                )
                return
            warnings.warn(
                "STORAGE_PROFILES is deprecated. Please use KEG_STORAGE_PROFILES instead.",
                DeprecationWarning,
            )
            app.config["KEG_STORAGE_PROFILES"] = app.config.pop("STORAGE_PROFILES")

    def init_app(self, app: flask.Flask):
        self._migrate_config(app)
        app.storage = self

        self._interfaces = collections.OrderedDict(
            (params['name'], interface(**params))
            for interface, params in app.config['KEG_STORAGE_PROFILES']
        )
        self.interface = next(iter(self._interfaces)) if self._interfaces else None

        self.init_cli(app)

    def get_interface(self, interface: Optional[str] = None):
        interface = interface or self.interface
        if interface is None:
            raise ValueError("no interface was specified")
        elif interface not in self._interfaces:
            raise ValueError(f"invalid interface '{interface}'")
        return self._interfaces[interface]

    def init_cli(self, app: flask.Flask) -> None:
        cli.add_cli_to_app(app, self.cli_group_name)


def _disable_csrf(func):
    if 'csrf' in flask.current_app.extensions:
        return flask.current_app.extensions['csrf'].exempt(func)
    return func


class LinkViewMixin:
    decorators = (_disable_csrf,)

    def get_storage_backend(self) -> backends.InternalLinksStorageBackend:
        raise NotImplementedError

    def get_request_token(self):
        return flask.request.args.get('token')

    def get_token_data(
            self,
            storage: backends.InternalLinksStorageBackend
    ) -> backends.InternalLinkTokenData:
        token = self.get_request_token()
        if not token:
            flask.abort(404)

        try:
            return storage.deserialize_link_token(token=token)
        except itsdangerous.BadData:
            flask.abort(403)

    def get(self):
        storage = self.get_storage_backend()
        token_data = self.get_token_data(storage)

        if not token_data.allow_download:
            flask.abort(403)

        fp = storage.open(token_data.path, backends.FileMode.read)
        return flask.Response(
            fp.iter_chunks(),
            mimetype='application/octet-stream',
        )

    def on_upload_success(self, token_data: backends.InternalLinkTokenData):
        return flask.Response('OK', status=200)

    def post(self):
        storage = self.get_storage_backend()
        token_data = self.get_token_data(storage)

        if not token_data.allow_upload:
            flask.abort(403)

        storage.upload(flask.request.stream, token_data.path)
        return self.on_upload_success(token_data)

    def put(self):
        return self.post()

    def on_delete_success(self, token_data: backends.InternalLinkTokenData):
        return flask.Response('OK', status=200)

    def delete(self):
        storage = self.get_storage_backend()
        token_data = self.get_token_data(storage)

        if not token_data.allow_remove:
            flask.abort(403)
        storage.delete(token_data.path)
        return self.on_delete_success(token_data)



from keg_storage.version import VERSION as __version__  # noqa
from importlib.util import find_spec as _find_spec

if _find_spec('keg'):
    from keg_storage.plugin import (  # noqa: F401
        Storage,
        LinkViewMixin,
    )
from keg_storage.backends import *  # noqa



import os
import string
from operator import attrgetter
from typing import (
    List,
    Optional,
    Union,
)

import pathlib

import arrow

from keg_storage.backends import base


class LocalFSError(Exception):
    pass


class LocalFSFile(base.RemoteFile):
    def __init__(self, path: pathlib.Path, mode: base.FileMode):
        self.path = path
        if mode & base.FileMode.read and mode & base.FileMode.write:
            str_mode = 'r+b'
        else:
            str_mode = str(mode)
        self.fp = path.open(str_mode)
        super().__init__(mode)

    def read(self, size: int) -> bytes:
        return self.fp.read(size)

    def write(self, data: bytes) -> None:
        self.fp.write(data)

    def close(self):
        self.fp.close()


class LocalFSStorage(base.InternalLinksStorageBackend):
    disallowed_path_chars = frozenset(set('~?*' + string.whitespace) - {' '})

    def __init__(
            self,
            root: Union[str, pathlib.Path],
            linked_endpoint: Optional[str] = None,
            secret_key: Optional[bytes] = None,
            name: str = None
    ):
        self.root = pathlib.Path(root).resolve()
        if not self.root.is_dir():
            raise LocalFSError('Storage root does not exist or is not a directory')

        super().__init__(
            linked_endpoint=linked_endpoint,
            secret_key=secret_key,
            name=name if name is not None else f'fs-{root.name}'
        )

    def _is_under_root(self, path: pathlib.Path) -> bool:
        try:
            path.resolve().relative_to(self.root)
        except ValueError:
            return False
        return True

    def _resolve_path(self, path: str) -> pathlib.Path:
        return self.root.joinpath(path).resolve()

    def _is_file(self, path: pathlib.Path) -> bool:
        return path.is_file() and not path.is_symlink()

    def _validate_path(self, path: str):
        if any(c in self.disallowed_path_chars for c in path):
            raise ValueError('Unsupported characters in path')

    def list(self, path: str) -> List[base.ListEntry]:
        self._validate_path(path)
        resolved_path = self._resolve_path(path)
        if not self._is_under_root(resolved_path):
            # prevent escaping the root by including `..` or symlinks in `path`
            raise LocalFSError('Invalid path')

        if not resolved_path.is_dir():
            raise LocalFSError(f'{path} does not exist or is not a directory')

        lst = []
        for dir, _, files in os.walk(resolved_path, followlinks=False):
            for f in files:
                full_path = pathlib.Path(dir).joinpath(f)
                if not self._is_file(full_path):
                    # Skip symlinks and special files like sockets
                    continue

                full_path = full_path.resolve()
                if not self._is_under_root(full_path):
                    continue

                st = full_path.stat()
                rel_path = full_path.relative_to(self.root)

                lst.append(base.ListEntry(
                    name=str(rel_path),
                    size=st.st_size,
                    last_modified=arrow.get(st.st_mtime)
                ))
        return sorted(lst, key=attrgetter('name'))

    def open(self, path: str, mode: Union[base.FileMode, str]):
        self._validate_path(path)
        path = self._resolve_path(path)
        if not self._is_under_root(path):
            raise LocalFSError('Invalid path')
        if path.exists() and not self._is_file(path):
            raise LocalFSError('Invalid path')

        mode = base.FileMode.as_mode(mode)

        if mode & base.FileMode.write:
            path.parent.mkdir(parents=True, exist_ok=True)

        return LocalFSFile(path, mode)

    def copy(self, path: str, new_path: str):
        return self.put(path, new_path)

    def delete(self, path: str):
        self._validate_path(path)
        path = self._resolve_path(path)
        if not self._is_under_root(path):
            raise LocalFSError('Invalid path')
        if not path.exists():
            return

        if not self._is_file(path):
            raise LocalFSError('Invalid path')

        path.unlink()



import io
import os
import pathlib
import re
from unittest import mock
from urllib.parse import urlparse

import arrow
import click
import freezegun
import itsdangerous
import pytest

import keg_storage
from keg_storage.backends.base import (
    InternalLinkTokenData,
    ListEntry,
    FileMode,
    RemoteFile,
    ShareLinkOperation,
)
from keg_storage.cli import handle_not_found


class FakeRemoteFile(RemoteFile):
    iter_chunk_size = 20

    def __init__(self, path: pathlib.Path, mode: FileMode):
        super().__init__(mode)
        self.file = path.open(mode=str(self.mode))

    def read(self, size):
        return self.file.read(size)

    def write(self, data):
        return self.file.write(data)

    def close(self):
        self.file.close()


class FakeBackend(keg_storage.InternalLinksStorageBackend):
    def __init__(
            self,
            base_dir: pathlib.Path,
            secret_key: bytes = None,
            linked_endpoint: str = None
    ):
        super().__init__(secret_key=secret_key, linked_endpoint=linked_endpoint, name="fake")
        self.base_dir = base_dir

    def list(self, path: str):
        results = []
        for dirent in os.scandir(self.base_dir / path):
            stat = dirent.stat()
            results.append(
                ListEntry(
                    name=dirent.name,
                    last_modified=arrow.get(stat.st_mtime),
                    size=stat.st_size
                )
            )
        return results

    def open(self, path, mode):
        path = os.path.join(self.base_dir, path)
        return FakeRemoteFile(self.base_dir / path, mode)

    def delete(self, path):
        path = os.path.join(self.base_dir, path)
        os.unlink(path)


class TestStorageBackend:

    def test_methods_not_implemented(self):
        interface = keg_storage.StorageBackend("incomplete")

        cases = {
            interface.list: ("path",),
            interface.delete: ("path",),
            interface.open: ("path", FileMode.read),
        }

        for method, args in cases.items():
            with pytest.raises(NotImplementedError):
                method(*args)

    def test_get(self, tmp_path: pathlib.Path):
        remote = tmp_path / 'remote'
        local = tmp_path / 'local'

        remote.mkdir()
        local.mkdir()

        data = b'a' * 15000000
        with (remote / 'input_file.txt').open('wb') as fp:
            fp.write(data)

        interface = FakeBackend(remote)

        output_path = local / 'output_file.txt'
        interface.get('input_file.txt', str(output_path))

        with output_path.open('rb') as of:
            assert of.read() == data

    def test_download(self, tmp_path: pathlib.Path):
        remote = tmp_path / "remote"

        remote.mkdir()

        data = b"a" * 15_000_000
        with (remote / "input_file.txt").open("wb") as fp:
            fp.write(data)

        buf = io.BytesIO()

        interface = FakeBackend(remote)
        interface.download("input_file.txt", buf)

        assert buf.getvalue() == data

    def test_download_progress(self, tmp_path: pathlib.Path):
        progress_updates = []

        def progress_callback(n: int) -> None:
            progress_updates.append(n)

        remote = tmp_path / "remote"
        remote.mkdir()

        data = b"a" * 15_000_000
        with (remote / "input_file.txt").open("wb") as fp:
            fp.write(data)

        buf = io.BytesIO()

        interface = FakeBackend(remote)
        interface.download("input_file.txt", buf, progress_callback=progress_callback)

        assert len(progress_updates) > 0
        assert progress_updates[-1] == len(data)

        assert buf.getvalue() == data

    def test_put(self, tmp_path: pathlib.Path):
        remote = tmp_path / 'remote'
        local = tmp_path / 'local'

        remote.mkdir()
        local.mkdir()

        data = b'a' * 15000000
        input_path = local / 'input_file.txt'
        with input_path.open('wb') as fp:
            fp.write(data)

        interface = FakeBackend(remote)
        interface.put(str(input_path), 'output_file.txt')

        with (remote / 'output_file.txt').open('rb') as of:
            assert of.read() == data

    def test_upload(self, tmp_path: pathlib.Path):
        remote = tmp_path / "remote"

        remote.mkdir()

        data = b"a" * 15_000_000
        buf = io.BytesIO(data)

        interface = FakeBackend(remote)
        interface.upload(buf, "output_file.txt")

        with (remote / "output_file.txt").open("rb") as of:
            assert of.read() == data

    def test_upload_progress(self, tmp_path: pathlib.Path):
        progress_updates = []

        def progress_callback(n: int) -> None:
            progress_updates.append(n)

        remote = tmp_path / "remote"
        remote.mkdir()

        data = b"a" * 15_000_000
        buf = io.BytesIO(data)

        interface = FakeBackend(remote)
        interface.upload(buf, "output_file.txt", progress_callback=progress_callback)

        assert len(progress_updates) > 0
        assert progress_updates[-1] == len(data)

        with (remote / "output_file.txt").open("rb") as of:
            assert of.read() == data

    def test_str(self, tmp_path: pathlib.Path):
        interface = FakeBackend(tmp_path)
        assert str(interface) == 'FakeBackend'

    def test_remote_file_iter_chunks(self, tmp_path: pathlib.Path):
        file_path = tmp_path / 'test_file.txt'
        with file_path.open('wb') as fp:
            fp.write(b"a" * 100 + b"b" * 100 + b"c" * 5)

        file = FakeRemoteFile(file_path, FileMode.read)
        chunks = list(file.iter_chunks(100))
        assert chunks == [
            b'a' * 100,
            b'b' * 100,
            b'c' * 5
        ]

    def test_remote_file_closes_on_delete(self, tmp_path: pathlib.Path):
        file_path = tmp_path / 'test_file.txt'
        file = FakeRemoteFile(file_path, FileMode.write)

        real_file = file.file
        assert real_file.closed is False

        del file
        assert real_file.closed is True

    def test_remote_file_context_manager(self, tmp_path: pathlib.Path):
        file_path = tmp_path / 'test_file.txt'
        file = FakeRemoteFile(file_path, FileMode.write)
        real_file = file.file

        with file:
            assert real_file.closed is False
        assert real_file.closed is True

    def test_remote_file_iter(self, tmp_path: pathlib.Path):
        file_path = tmp_path / 'test_file.txt'
        with file_path.open('wb') as fp:
            fp.write(b"a" * 20 + b"b" * 20 + b"c" * 20)

        file = FakeRemoteFile(file_path, FileMode.read)
        chunks = list(file)
        assert chunks == [
            b'a' * 20,
            b'b' * 20,
            b'c' * 20
        ]


class TestInternalLinkTokenData:
    def test_serialize(self):
        tok_data = InternalLinkTokenData(path='foo', operations=ShareLinkOperation.upload)
        assert tok_data.serialize() == {'key': 'foo', 'op': 'u'}

    def test_deserialize(self):
        tok_data = InternalLinkTokenData.deserialize({'key': 'foo', 'op': 'u'})
        assert tok_data.path == 'foo'
        assert tok_data.operations == ShareLinkOperation.upload

    def test_allowed_operations(self):
        tok_data = InternalLinkTokenData(path='foo', operations=ShareLinkOperation.download)
        assert tok_data.allow_download is True
        assert tok_data.allow_upload is False
        assert tok_data.allow_remove is False

        tok_data = InternalLinkTokenData(path='foo', operations=ShareLinkOperation.upload)
        assert tok_data.allow_download is False
        assert tok_data.allow_upload is True
        assert tok_data.allow_remove is False

        tok_data = InternalLinkTokenData(path='foo', operations=ShareLinkOperation.remove)
        assert tok_data.allow_download is False
        assert tok_data.allow_upload is False
        assert tok_data.allow_remove is True

        tok_data = InternalLinkTokenData(
            path='foo',
            operations=ShareLinkOperation.download | ShareLinkOperation.upload
        )
        assert tok_data.allow_download is True
        assert tok_data.allow_upload is True
        assert tok_data.allow_remove is False

        tok_data = InternalLinkTokenData(
            path='foo',
            operations=ShareLinkOperation.download | ShareLinkOperation.remove
        )
        assert tok_data.allow_download is True
        assert tok_data.allow_upload is False
        assert tok_data.allow_remove is True


class TestInternalLinkStorageBackend:
    def test_create_link_token_no_secret_key(self, tmp_path: pathlib.Path):
        backend = FakeBackend(tmp_path)
        with pytest.raises(ValueError, match='Backend must be configured with secret_key .*'):
            backend.create_link_token(path='foo', operation=ShareLinkOperation.download,
                                      expire=arrow.utcnow())

    @freezegun.freeze_time('2020-04-27')
    def test_create_link_token_success(self, tmp_path: pathlib.Path):
        backend = FakeBackend(tmp_path, secret_key=b'a' * 32)
        token = backend.create_link_token(
            path='foo',
            operation=ShareLinkOperation.download,
            expire=arrow.get(2020, 4, 27, 1)
        )

        signer = itsdangerous.TimedJSONWebSignatureSerializer(
            secret_key=b'a' * 32,
            salt='fake'
        )
        body, header = signer.loads(token, salt='fake', return_header=True)
        assert body == {
            'key': 'foo',
            'op': 'd'
        }
        assert header == {
            'alg': 'HS512',
            'iat': arrow.get(2020, 4, 27).timestamp(),
            'exp': arrow.get(2020, 4, 27, 1).timestamp()
        }

    def test_deserialize_link_token_no_secret_key(self, tmp_path: pathlib.Path):
        backend = FakeBackend(tmp_path)
        with pytest.raises(ValueError, match='Backend must be configured with secret_key .*'):
            backend.deserialize_link_token('foo')

    @freezegun.freeze_time('2020-04-27')
    def test_deserialize_link_token_bad_signature(self, tmp_path: pathlib.Path):
        backend = FakeBackend(tmp_path, secret_key=b'a' * 32)
        token = backend.create_link_token(path='foo', operation=ShareLinkOperation.download,
                                          expire=arrow.get(2020, 4, 27, 1))
        backend.name = 'fake1'
        with pytest.raises(itsdangerous.BadSignature, match='Signature .* does not match'):
            backend.deserialize_link_token(token)

    def test_deserialize_link_token_expired(self, tmp_path: pathlib.Path):
        backend = FakeBackend(tmp_path, secret_key=b'a' * 32)
        with freezegun.freeze_time('2020-04-26'):
            token = backend.create_link_token(path='foo', operation=ShareLinkOperation.download,
                                              expire=arrow.get(2020, 4, 26, 23, 59, 59))

        with freezegun.freeze_time('2020-04-27'):
            with pytest.raises(itsdangerous.SignatureExpired, match='Signature expired'):
                backend.deserialize_link_token(token)

    @freezegun.freeze_time('2020-04-27')
    def test_deserialize_link_token_success(self, tmp_path: pathlib.Path):
        backend = FakeBackend(tmp_path, secret_key=b'a' * 32)
        token = backend.create_link_token(
            path='foo',
            operation=ShareLinkOperation.download,
            expire=arrow.get(2020, 4, 27, 1)
        )

        result = backend.deserialize_link_token(token)
        assert result.path == 'foo'
        assert result.operations == ShareLinkOperation.download

    @freezegun.freeze_time('2020-04-27')
    def test_link_to_no_secret_key(self, tmp_path: pathlib.Path):
        backend = FakeBackend(tmp_path, linked_endpoint='aaa.xyz')
        with pytest.raises(
                ValueError,
                match='Backend must be configured with linked_endpoint and secret_key .*'
        ):
            backend.link_to(
                path='foo',
                operation=ShareLinkOperation.download,
                expire=arrow.get(2020, 4, 28)
            )

    @freezegun.freeze_time('2020-04-27')
    def test_link_to_no_linked_endpoint(self, tmp_path: pathlib.Path):
        backend = FakeBackend(tmp_path, secret_key=b'a' * 32)
        with pytest.raises(
                ValueError,
                match='Backend must be configured with linked_endpoint and secret_key .*'
        ):
            backend.link_to(
                path='foo',
                operation=ShareLinkOperation.download,
                expire=arrow.get(2020, 4, 28)
            )

    @freezegun.freeze_time('2020-04-27')
    @mock.patch('flask.url_for', autospec=True, spec_set=True)
    def test_link_to_success(self, m_url_for, tmp_path: pathlib.Path):
        m_url_for.side_effect = 'http://localhost/storage?token={token}'.format
        backend = FakeBackend(tmp_path, secret_key=b'a' * 32, linked_endpoint='aaa.xyz')
        link = backend.link_to(
            path='foo',
            operation=ShareLinkOperation.download,
            expire=arrow.get(2020, 4, 28)
        )
        query = urlparse(link).query
        assert re.match(r'^token=[\w\-_]+\.[\w\-_]+\.[\w\-_]+$', query)


class TestFileNotFoundException:
    def test_click_wrapper(self, tmp_path: pathlib.Path):
        backend = FakeBackend(tmp_path)

        @handle_not_found
        def test_func():
            raise keg_storage.FileNotFoundInStorageError(backend, 'foo')

        with pytest.raises(click.FileError) as exc_info:
            test_func()
        assert exc_info.value.filename == 'foo'
        assert exc_info.value.message == 'Not found in FakeBackend.'


class TestFileMode:
    def test_str(self):
        assert str(FileMode.read) == 'rb'
        assert str(FileMode.write) == 'wb'
        assert str(FileMode.read | FileMode.write) == 'rwb'

    def test_as_mode(self):
        assert FileMode.as_mode(FileMode.read) == FileMode.read

        assert FileMode.as_mode("r") == FileMode.read
        assert FileMode.as_mode("w") == FileMode.write
        assert FileMode.as_mode("rw") == FileMode.read | FileMode.write
        assert FileMode.as_mode("wr") == FileMode.read | FileMode.write
        assert FileMode.as_mode("rb") == FileMode.read

        with pytest.raises(
            ValueError, match=re.escape("as_mode() accepts only FileMode or str arguments")
        ):
            FileMode.as_mode(1)


class TestShareLinkOperation:
    def test_str(self):
        assert str(ShareLinkOperation.download) == 'd'
        assert str(ShareLinkOperation.upload) == 'u'
        assert str(ShareLinkOperation.remove) == 'r'
        assert str(ShareLinkOperation.download | ShareLinkOperation.upload) == 'du'
        assert str(ShareLinkOperation.download | ShareLinkOperation.remove) == 'dr'
        assert str(ShareLinkOperation.upload | ShareLinkOperation.remove) == 'ur'
        assert str(
            ShareLinkOperation.download | ShareLinkOperation.upload | ShareLinkOperation.remove
        ) == 'dur'

    def test_as_operation(self):
        assert (
            ShareLinkOperation.as_operation(ShareLinkOperation.download) ==
            ShareLinkOperation.download
        )
        assert (
            ShareLinkOperation.as_operation(ShareLinkOperation.download | ShareLinkOperation.upload)
            == ShareLinkOperation.download | ShareLinkOperation.upload
        )

        assert ShareLinkOperation.as_operation('d') == ShareLinkOperation.download
        assert ShareLinkOperation.as_operation('u') == ShareLinkOperation.upload
        assert ShareLinkOperation.as_operation('r') == ShareLinkOperation.remove
        assert (
            ShareLinkOperation.as_operation('du') ==
            ShareLinkOperation.download | ShareLinkOperation.upload
        )
        assert (
            ShareLinkOperation.as_operation('ud') ==
            ShareLinkOperation.download | ShareLinkOperation.upload
        )
        assert (
            ShareLinkOperation.as_operation('dr') ==
            ShareLinkOperation.download | ShareLinkOperation.remove
        )
        assert (
            ShareLinkOperation.as_operation('rd') ==
            ShareLinkOperation.download | ShareLinkOperation.remove
        )
        assert (
            ShareLinkOperation.as_operation('ur') ==
            ShareLinkOperation.upload | ShareLinkOperation.remove
        )
        assert (
            ShareLinkOperation.as_operation('ru') ==
            ShareLinkOperation.upload | ShareLinkOperation.remove
        )
        assert (
            ShareLinkOperation.as_operation('dur') ==
            ShareLinkOperation.download | ShareLinkOperation.upload | ShareLinkOperation.remove
        )
        assert (
            ShareLinkOperation.as_operation('rud') ==
            ShareLinkOperation.download | ShareLinkOperation.upload | ShareLinkOperation.remove
        )

        with pytest.raises(
            ValueError,
            match=re.escape("as_operation() accepts only ShareLinkOperation or str arguments")
        ):
            ShareLinkOperation.as_operation(1)



import os
import pathlib
import random
from typing import Callable

import pytest
from blazeutils import randchars

from keg_storage import backends


class TestLocalFSStorage:
    def test_init(self, tmp_path: pathlib.Path):
        root = tmp_path.joinpath('real')
        root.mkdir()
        tmp_path.joinpath('link').symlink_to(root, target_is_directory=True)

        fs = backends.LocalFSStorage(root)
        assert fs.root == tmp_path / 'real'
        assert fs.name == 'fs-real'

        fs = backends.LocalFSStorage(tmp_path / 'link')
        assert fs.root == tmp_path / 'real'
        assert fs.name == 'fs-link'

    def test_init_not_dir(self, tmp_path: pathlib.Path):
        root = tmp_path.joinpath('foo')
        root.touch()

        with pytest.raises(backends.LocalFSError) as exc:
            backends.LocalFSStorage(root)
        assert str(exc.value) == 'Storage root does not exist or is not a directory'

    def test_init_does_not_exist(self, tmp_path: pathlib.Path):
        with pytest.raises(backends.LocalFSError) as exc:
            backends.LocalFSStorage(tmp_path / 'foo')
        assert str(exc.value) == 'Storage root does not exist or is not a directory'

    def test_list_success(self, tmp_path: pathlib.Path):
        root = tmp_path.joinpath('root')

        dir1 = root.joinpath('dir1')
        dir2 = dir1.joinpath('dir2')
        dir2.mkdir(parents=True)

        dir3 = root.joinpath('dir3')
        dir3.mkdir()

        # empty dir should be skipped
        root.joinpath('dir4').mkdir()

        # create some files that will be listed
        dir1.joinpath('file1.txt').touch()
        dir1.joinpath('file2.txt').touch()
        dir2.joinpath('file1.txt').touch()
        with dir3.joinpath('file3.txt').open('wb') as fp:
            fp.write(os.urandom(100))

        # create links that should be skipped
        dir3.joinpath('link1').symlink_to(dir2, target_is_directory=True)
        dir3.joinpath('link2.txt').symlink_to(dir2 / 'file1.txt', target_is_directory=False)

        # create a FIFO that should be skipped
        os.mkfifo(dir3 / 'pipe')

        # create an link to an external directory that should be skipped
        external = tmp_path.joinpath('external')
        external.mkdir()
        dir3.joinpath('link3').symlink_to(external, target_is_directory=True)

        fs = backends.LocalFSStorage(root)
        results = fs.list('')

        assert len(results) == 4
        assert results[0].name == 'dir1/dir2/file1.txt'
        assert results[0].size == 0

        assert results[1].name == 'dir1/file1.txt'
        assert results[1].size == 0

        assert results[2].name == 'dir1/file2.txt'
        assert results[2].size == 0

        assert results[3].name == 'dir3/file3.txt'
        assert results[3].size == 100

        results = fs.list('dir1')
        assert len(results) == 3
        assert results[0].name == 'dir1/dir2/file1.txt'
        assert results[0].size == 0

        assert results[1].name == 'dir1/file1.txt'
        assert results[1].size == 0

        assert results[2].name == 'dir1/file2.txt'
        assert results[2].size == 0

        results = fs.list('dir1/dir2')
        assert len(results) == 1
        assert results[0].name == 'dir1/dir2/file1.txt'
        assert results[0].size == 0

    def test_list_failures(self, tmp_path: pathlib.Path):
        root = tmp_path.joinpath('root')

        ext1 = tmp_path.joinpath('ext1')
        ext2 = ext1.joinpath('ext2')
        ext2.mkdir(parents=True)
        ext2.joinpath('file1.txt').touch()

        dir1 = root.joinpath('dir1')
        dir1.mkdir(parents=True)
        dir1.joinpath('file2.txt').touch()

        link = dir1.joinpath('link')
        link.symlink_to(ext1, target_is_directory=True)

        fs = backends.LocalFSStorage(root)

        # Attempt to break sandbox with a link
        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.list('dir1/link')

        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.list('dir1/link/ext2')

        # Attempt to break sandbox with a relative path
        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.list('dir1/../../')

        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.list('dir1/../../external')

        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.list('/tmp')

        with pytest.raises(ValueError, match='Unsupported characters in path'):
            fs.list('~/')

        # Attempt to list a file
        with pytest.raises(backends.LocalFSError,
                           match='dir1/file2.txt does not exist or is not a directory'):
            fs.list('dir1/file2.txt')

        # Attempt to list a directory that does not exist
        with pytest.raises(backends.LocalFSError,
                           match='dir2 does not exist or is not a directory'):
            fs.list('dir2')

    def test_open_for_reading(self, tmp_path: pathlib.Path):
        root = tmp_path.joinpath('root')
        root.mkdir()
        file_path = root.joinpath('file.txt')
        file_data = os.urandom(100)
        with file_path.open('wb') as fp:
            fp.write(file_data)

        fs = backends.LocalFSStorage(root)

        with fs.open('file.txt', backends.FileMode.read) as f:
            assert f.read(10) == file_data[:10]
            assert f.read(100) == file_data[10:]

        assert f.fp.closed is True

        with fs.open('file.txt', 'r') as f:
            assert f.read(100) == file_data

        assert f.fp.closed is True

    def test_open_for_writing(self, tmp_path: pathlib.Path):
        root = tmp_path.joinpath('root')
        root.mkdir()
        file_path = root.joinpath('file.txt')

        fs = backends.LocalFSStorage(root)

        # File does not exist yet
        file_data1 = os.urandom(100)
        with fs.open('file.txt', backends.FileMode.write) as f:
            f.write(file_data1)

        with file_path.open('rb') as fp:
            assert fp.read() == file_data1
        assert f.fp.closed is True

        # Existing file
        file_data2 = os.urandom(200)
        with fs.open('file.txt', 'w') as f:
            f.write(file_data2)

        with file_path.open('rb') as fp:
            assert fp.read() == file_data2
        assert f.fp.closed is True

    def test_copy_file(self, tmp_path: pathlib.Path):
        root = tmp_path.joinpath('root')
        root.mkdir()

        file_path = root.joinpath('file.txt')
        file_path_copy = root.joinpath('file2.txt')
        fs = backends.LocalFSStorage(root)
        file_data1 = os.urandom(100)

        with fs.open('file.txt', backends.FileMode.write) as f:
            f.write(file_data1)
        fs.copy(str(file_path), str(file_path_copy))

        with file_path_copy.open('rb') as fp:
            assert fp.read() == file_data1

    def test_open_for_writing_creates_directories(self, tmp_path: pathlib.Path):
        root = tmp_path.joinpath('root')
        root.mkdir()

        fs = backends.LocalFSStorage(root)
        with fs.open('dir1/dir2/file.txt', backends.FileMode.write):
            pass

        assert root.joinpath('dir1', 'dir2').is_dir()

    def test_open_for_read_write(self, tmp_path: pathlib.Path):
        root = tmp_path.joinpath('root')
        root.mkdir()
        file_path = root.joinpath('file.txt')
        file_data = os.urandom(100)
        with file_path.open('wb') as fp:
            fp.write(file_data)

        fs = backends.LocalFSStorage(root)

        new_data = os.urandom(100)
        with fs.open('file.txt', backends.FileMode.read | backends.FileMode.write) as f:
            assert f.read(10) == file_data[:10]
            f.write(new_data)

        with file_path.open('rb') as fp:
            assert fp.read() == file_data[:10] + new_data
        assert f.fp.closed is True

    def test_open_failures(self, tmp_path: pathlib.Path):
        root = tmp_path.joinpath('root')
        root.mkdir()
        root.joinpath('rootfile.txt').touch()

        ext1 = tmp_path.joinpath('ext1')
        ext1.mkdir(parents=True)
        ext1.joinpath('file1.txt').touch()

        dir1 = root.joinpath('dir1')
        dir1.mkdir(parents=True)
        dir1.joinpath('file2.txt').touch()

        link1 = dir1.joinpath('link')
        link1.symlink_to(ext1, target_is_directory=True)

        link2 = dir1.joinpath('link.txt')
        link2.symlink_to(ext1 / 'file1.txt', target_is_directory=False)

        fs = backends.LocalFSStorage(root)

        # Attempt to break sandbox with a link
        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.open('dir1/link', 'r')

        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.open('dir1/link.txt', 'r')

        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.open('dir1/link/file1.txt', 'r')

        # Attempt to break sandbox with a relative path
        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.open('dir1/../../rootfile.txt', 'r')

        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.open('dir1/../../external/file1.txt', 'r')

        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.open('/tmp/file.txt', 'w')

        with pytest.raises(ValueError):
            fs.open('~/file.txt', 'w')

        # Attempt to open non-files
        os.mkfifo(root / 'fifo')
        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.open('fifo', 'r')

        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.open('dir1', 'r')

    def test_delete_success(self, tmp_path: pathlib.Path):
        root = tmp_path.joinpath('root')
        root.mkdir()
        file_path = root.joinpath('file.txt')
        file_path.touch()

        fs = backends.LocalFSStorage(root)

        assert file_path.exists()
        fs.delete('file.txt')
        assert not file_path.exists()

        # None existant file should return early
        fs.delete('file2.txt')

    def test_delete_failures(self, tmp_path: pathlib.Path):
        root = tmp_path.joinpath('root')
        root.mkdir()
        root.joinpath('rootfile.txt').touch()

        ext1 = tmp_path.joinpath('ext1')
        ext1.mkdir(parents=True)
        ext1.joinpath('file1.txt').touch()

        dir1 = root.joinpath('dir1')
        dir1.mkdir(parents=True)
        dir1.joinpath('file2.txt').touch()

        link1 = dir1.joinpath('link')
        link1.symlink_to(ext1, target_is_directory=True)

        link2 = dir1.joinpath('link.txt')
        link2.symlink_to(ext1 / 'file1.txt', target_is_directory=False)

        fs = backends.LocalFSStorage(root)

        # Attempt to break sandbox with a link
        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.delete('dir1/link')
        assert link1.exists()

        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.delete('dir1/link.txt')
        assert link2.exists()

        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.delete('dir1/link/file1.txt')
        assert ext1.joinpath('file1.txt').exists()

        # Attempt to break sandbox with a relative path
        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.delete('dir1/../../rootfile.txt')
        assert root.joinpath('rootfile.txt').exists()

        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.delete('dir1/../../external/file1.txt')
        assert ext1.joinpath('file1.txt').exists()

        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.delete('/tmp/file.txt')

        with pytest.raises(ValueError):
            fs.delete('~/file.txt')

        # Attempt to delete non-files
        os.mkfifo(root / 'fifo')
        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.delete('fifo')
        assert root.joinpath('fifo').exists()

        with pytest.raises(backends.LocalFSError, match='Invalid path'):
            fs.open('dir1', 'r')
        assert dir1.exists()

    @pytest.mark.parametrize('method,args', [
        (backends.LocalFSStorage.list, tuple()),
        (backends.LocalFSStorage.open, ('w',)),
        (backends.LocalFSStorage.delete, tuple()),
    ])
    @pytest.mark.parametrize('char', ['~', '?', '*', '\t', '\n', '\r', '\x0b', '\x0c'])
    def test_validates_path(self, method: Callable, args: tuple, char: str, tmp_path: pathlib.Path):
        fs = backends.LocalFSStorage(tmp_path)
        chars = list(randchars()) + [char]
        random.shuffle(chars)
        file_name = ''.join(chars)

        with pytest.raises(ValueError, match='Unsupported characters in path'):
            method(fs, file_name, *args)



import base64
import datetime
import re
import string
import urllib.parse as urlparse
from io import BytesIO
from typing import Union
from unittest import mock
import urllib.parse

import arrow
import pytest
from azure.storage.blob import BlobClient, BlobProperties, ContainerClient, BlobPrefix

from keg_storage import backends
from keg_storage.backends import base


def create_storage(**kwargs):
    return backends.AzureStorage(**{
        'account': 'foo',
        'key': base64.b64encode(b'a' * 64).decode(),
        'bucket': 'test',
        **kwargs,
    })


class TestAzureStorageBasics:
    def test_construct_incomplete(self):
        with pytest.raises(
            ValueError,
            match=(
                "Must provide a sas_container_url, a sas_blob_url, "
                "or a combination of account, key, and bucket"
            ),
        ):
            backends.AzureStorage(account="foo", bucket="test")

    @mock.patch.object(ContainerClient, "from_container_url")
    def test_construct_with_container_url(self, m_from_container_url: mock.MagicMock):
        storage = backends.AzureStorage(
            sas_container_url="https://foo.blob.core.windows.net/test?sp=rwdl"
        )
        storage.upload(BytesIO(b"hello"), "inbox/message.txt")

        m_from_container_url.return_value.get_blob_client.assert_called_once_with(
            "inbox/message.txt"
        )

    @mock.patch.object(BlobClient, "from_blob_url")
    def test_construct_with_blob_url(self, m_from_blob_url: mock.MagicMock):
        blob_name = "inbox/message.txt"
        escaped_blob_name = urllib.parse.quote("inbox/message.txt", safe="")
        sas_blob_url = f"https://foo.blob.core.windows.net/test/{escaped_blob_name}?sp=cw"

        # Set this attribute on the mock client, because it is used for path validation.
        m_from_blob_url.return_value.blob_name = blob_name

        storage = backends.AzureStorage(sas_blob_url=sas_blob_url)
        storage.upload(BytesIO(b"hello"), blob_name)

        m_from_blob_url.assert_called_once_with(sas_blob_url)

        # Make sure path validation is working.
        with pytest.raises(ValueError, match="Invalid path for the configured SAS blob URL"):
            storage.upload(BytesIO(b"hello"), "inbox/another-message.txt")

        # Make sure list operation fails.
        with pytest.raises(ValueError, match="Cannot perform list operation .* SAS blob URL"):
            storage.list("inbox/")

    def test_open_read_write(self):
        storage = create_storage()
        with pytest.raises(
            NotImplementedError,
            match=re.escape("Read+write mode not supported by the Azure backend"),
        ):
            storage.open("foo", base.FileMode.read | base.FileMode.write)

    def test_open_bad_mode(self):
        storage = create_storage()
        with pytest.raises(
            ValueError,
            match=re.escape("Unsupported mode. Accepted modes are FileMode.read or FileMode.write"),
        ):
            storage.open("foo", base.FileMode(0))


@mock.patch.object(backends.AzureStorage, '_create_blob_client')
class TestAzureStorageOperations:
    @mock.patch.object(backends.AzureStorage, '_create_container_client')
    def test_list(self, m_container_client: mock.MagicMock, m_blob_client: mock.MagicMock):
        m_walk = m_container_client.return_value.walk_blobs

        def blob(name, last_modified, size):
            return BlobProperties(**{
                'name': name,
                'Last-Modified': last_modified,
                'Content-Length': size
            })

        prefix = BlobPrefix(prefix='abc/')

        m_walk.return_value = [
            prefix,
            blob('abc/foo.txt', datetime.datetime(2019, 10, 5, 1, 1, 1), 123),
            blob('abc/bar.txt', datetime.datetime(2019, 10, 5, 1, 1, 2), 321),
            blob('baz.txt', datetime.datetime(2019, 10, 5, 1, 1, 3), 100),
        ]
        storage = create_storage()
        results = storage.list('xyz')

        m_walk.assert_called_once_with('xyz/')
        assert results == [
            base.ListEntry(name='abc/', last_modified=None, size=0),
            base.ListEntry(name='abc/foo.txt', last_modified=arrow.get(2019, 10, 5, 1, 1, 1), size=123),  # noqa
            base.ListEntry(name='abc/bar.txt', last_modified=arrow.get(2019, 10, 5, 1, 1, 2), size=321),  # noqa
            base.ListEntry(name='baz.txt', last_modified=arrow.get(2019, 10, 5, 1, 1, 3), size=100),
        ]

    @mock.patch.object(backends.AzureStorage, '_create_container_client')
    def test_list_leading_slashes(
        self, m_container_client: mock.MagicMock, m_blob_client: mock.MagicMock
    ):
        m_walk = m_container_client.return_value.walk_blobs
        m_walk.return_value = []
        storage = create_storage()
        storage.list('//xyz')
        m_walk.assert_called_once_with('xyz/')

    def test_read_operations(self, m_blob_client: mock.MagicMock):
        storage = create_storage(chunk_size=10)

        data = ''.join([
            string.digits,
            string.ascii_lowercase,
            string.ascii_uppercase,
        ]).encode()
        chunks = [data[i : i + 10] for i in range(0, len(data), 10)]  # noqa: E203

        m_stream = m_blob_client.return_value.download_blob
        m_stream.return_value.chunks.return_value = iter(chunks)

        with storage.open('foo', base.FileMode.read) as f:
            assert f.read(1) == b'0'
            assert f.read(2) == b'12'
            assert f.read(10) == b'3456789abc'
            assert f.read(30) == b'defghijklmnopqrstuvwxyzABCDEFG'
            assert f.read(30) == b'HIJKLMNOPQRSTUVWXYZ'

    @mock.patch('keg_storage.backends.azure.os.urandom', autospec=True, spec_set=True)
    def test_write_operations(self, m_urandom: mock.MagicMock, m_blob_client: mock.MagicMock):
        storage = create_storage(chunk_size=10)

        m_urandom.side_effect = lambda x: b'\x00' * x

        block_data = {}
        blob = BytesIO()

        def mock_stage_block(**kwargs):
            block_id = kwargs['block_id']
            assert block_id not in block_data
            block_data[block_id] = kwargs['data']

        def mock_commit_block_list(**kwargs):
            blocks = kwargs['block_list']
            for b in blocks:
                blob.write(block_data[b.id])

        m_stage_block = m_blob_client.return_value.stage_block
        m_stage_block.side_effect = mock_stage_block

        m_commit_list = m_blob_client.return_value.commit_block_list
        m_commit_list.side_effect = mock_commit_block_list

        def block_id(index_bytes):
            return base64.b64encode(index_bytes + bytes([0] * 40)).decode()

        with storage.open('foo', base.FileMode.write) as f:
            f.write(b'ab')
            m_stage_block.assert_not_called()
            m_commit_list.assert_not_called()
            assert blob.getvalue() == b''
            assert block_data == {}

            f.write(b'cdefghijklm')
            m_stage_block.assert_called_once_with(
                block_id=block_id(b'\x00\x00\x00\x00\x00\x00\x00\x00'),
                data=b'abcdefghij',
            )
            m_commit_list.assert_not_called()
            assert blob.getvalue() == b''
            assert set(block_data.values()) == {b'abcdefghij'}

            f.write(b'nopqrstuvwxyz')
            m_stage_block.assert_called_with(
                block_id=block_id(b'\x00\x00\x00\x00\x00\x00\x00\x01'),
                data=b'klmnopqrst',
            )
            m_stage_block.reset_mock()
            m_commit_list.assert_not_called()
            assert blob.getvalue() == b''
            assert set(block_data.values()) == {b'abcdefghij', b'klmnopqrst'}

            f.write(b'12')
            m_stage_block.assert_not_called()
            m_commit_list.assert_not_called()
            assert blob.getvalue() == b''
            assert set(block_data.values()) == {b'abcdefghij', b'klmnopqrst'}

        m_stage_block.assert_called_with(
            block_id=block_id(b'\x00\x00\x00\x00\x00\x00\x00\x02'),
            data=b'uvwxyz12',
        )
        m_commit_list.assert_called_once()
        _, kwargs = m_commit_list.call_args
        assert [b.id for b in kwargs['block_list']] == [
            block_id(b'\x00\x00\x00\x00\x00\x00\x00\x00'),
            block_id(b'\x00\x00\x00\x00\x00\x00\x00\x01'),
            block_id(b'\x00\x00\x00\x00\x00\x00\x00\x02')
        ]
        assert blob.getvalue() == b'abcdefghijklmnopqrstuvwxyz12'

    def test_write_nothing(self, m_blob_client: mock.MagicMock):
        storage = create_storage()

        m_stage_block = m_blob_client.return_value.stage_block
        m_commit_list = m_blob_client.return_value.commit_block_list

        f = storage.open('foo', base.FileMode.write)
        f.close()

        m_stage_block.assert_not_called()
        m_commit_list.assert_not_called()

    def test_open_leading_slashes(self, m_blob_client: mock.MagicMock):
        storage = create_storage()
        storage.open("//xyz", "r")
        m_blob_client.assert_called_once_with("xyz")

    def test_delete(self, m_blob_client: mock.MagicMock):
        storage = create_storage()
        storage.delete("foo")

        m_blob_client.assert_called_once_with("foo")
        m_blob_client.return_value.delete_blob.assert_called_once_with()

    def test_delete_leading_slashes(self, m_blob_client: mock.MagicMock):
        storage = create_storage()
        storage.delete("/foo")

        m_blob_client.assert_called_once_with("foo")
        m_blob_client.return_value.delete_blob.assert_called_once_with()


class TestAzureStorageUtilities:
    @pytest.mark.parametrize('expire', [
        arrow.get(2019, 1, 2, 3, 4, 5),
        datetime.datetime(2019, 1, 2, 3, 4, 5)
    ])
    def test_upload_url(self, expire: Union[arrow.Arrow, datetime.datetime]):
        storage = create_storage()
        with pytest.warns(DeprecationWarning):
            url = storage.create_upload_url(
                'abc/def.txt',
                expire=expire
            )
        parsed = urlparse.urlparse(url)
        assert parsed.netloc == 'foo.blob.core.windows.net'
        assert parsed.path == '/test/abc%2Fdef.txt'
        qs = urlparse.parse_qs(parsed.query)

        assert qs['se'] == ['2019-01-02T03:04:05Z']
        assert qs['sp'] == ['cw']
        assert qs['sig']
        assert 'sip' not in qs

    @pytest.mark.parametrize('expire', [
        arrow.get(2019, 1, 2, 3, 4, 5),
        datetime.datetime(2019, 1, 2, 3, 4, 5)
    ])
    def test_download_url(self, expire: Union[arrow.Arrow, datetime.datetime]):
        storage = create_storage()
        with pytest.warns(DeprecationWarning):
            url = storage.create_download_url(
                'abc/def.txt',
                expire=expire
            )
        parsed = urlparse.urlparse(url)
        assert parsed.netloc == 'foo.blob.core.windows.net'
        assert parsed.path == '/test/abc%2Fdef.txt'
        qs = urlparse.parse_qs(parsed.query)

        assert qs['se'] == ['2019-01-02T03:04:05Z']
        assert qs['sp'] == ['r']
        assert qs['sig']
        assert 'sip' not in qs

    @pytest.mark.parametrize('expire', [
        arrow.get(2019, 1, 2, 3, 4, 5),
        datetime.datetime(2019, 1, 2, 3, 4, 5)
    ])
    @pytest.mark.parametrize('ops,perms', [
        (base.ShareLinkOperation.download, 'r'),
        (base.ShareLinkOperation.upload, 'cw'),
        (base.ShareLinkOperation.remove, 'd'),
        (base.ShareLinkOperation.download | base.ShareLinkOperation.upload, 'rcw'),
        (base.ShareLinkOperation.upload | base.ShareLinkOperation.remove, 'cwd'),
    ])
    def test_link_to(self, expire: Union[arrow.Arrow, datetime.datetime],
                     ops: base.ShareLinkOperation, perms: str):
        storage = create_storage()
        url = storage.link_to(
            'abc/def.txt',
            operation=ops,
            expire=expire
        )
        parsed = urlparse.urlparse(url)
        assert parsed.netloc == 'foo.blob.core.windows.net'
        assert parsed.path == '/test/abc%2Fdef.txt'
        qs = urlparse.parse_qs(parsed.query)

        assert qs['se'] == ['2019-01-02T03:04:05Z']
        assert qs['sp'] == [perms]
        assert qs['sig']
        assert 'sip' not in qs

    def test_sas_create_container_url(self):
        storage = backends.AzureStorage(
            **{"sas_container_url": "https://foo.blob.core.windows.net/test?sp=rwdl"}
        )

        with pytest.raises(ValueError, match="Cannot create a SAS URL without account credentials"):
            storage.create_container_url(arrow.get(2019, 1, 2, 3, 4, 5))

    @pytest.mark.parametrize('account,key', [
        (None, 'foo'),
        ('foo', None),
        (None, None),
    ])
    def test_sas_link_to(self, account, key):
        storage = backends.AzureStorage(
            account=account,
            key=key,
            sas_container_url='https://foo.blob.core.windows.net/test?sp=rwdl'
        )

        with pytest.raises(ValueError, match="Cannot create a SAS URL without account credentials"):
            storage.link_to("foo/bar.txt", backends.ShareLinkOperation.upload,
                            arrow.get(2019, 1, 2, 3, 4, 5))

    @pytest.mark.parametrize('expire', [
        arrow.get(2019, 1, 2, 3, 4, 5),
        datetime.datetime(2019, 1, 2, 3, 4, 5)
    ])
    def test_create_container_url(self, expire: Union[arrow.Arrow, datetime.datetime]):
        storage = create_storage()
        url = storage.create_container_url(expire=expire)
        parsed = urlparse.urlparse(url)
        assert parsed.netloc == 'foo.blob.core.windows.net'
        assert parsed.path == '/test'
        qs = urlparse.parse_qs(parsed.query)

        assert qs['se'] == ['2019-01-02T03:04:05Z']
        assert qs['sp'] == ['rwdl']
        assert qs['sig']
        assert 'sip' not in qs

        # with IP restriction
        url = storage.create_container_url(expire=expire, ip='127.0.0.1')
        parsed = urlparse.urlparse(url)
        assert parsed.netloc == 'foo.blob.core.windows.net'
        assert parsed.path == '/test'
        qs = urlparse.parse_qs(parsed.query)

        assert qs['se'] == ['2019-01-02T03:04:05Z']
        assert qs['sp'] == ['rwdl']
        assert qs['sig']
        assert qs['sip'] == ['127.0.0.1']



import contextlib
import io
import os
from datetime import date
from unittest import mock

import arrow
import click.testing
import freezegun
import pytest
from blazeutils.containers import LazyDict
from flask import current_app
from keg.testing import CLIBase, invoke_command

from keg_storage import FileNotFoundInStorageError, ShareLinkOperation


@mock.patch.object(current_app.storage, 'get_interface', autospec=True, spec_set=True)
class TestCLI(CLIBase):
    def test_no_location(self, m_get_interface):
        results = self.invoke('storage', 'list', 'foo/bar', exit_code=1)
        assert results.output == 'No location given and no default was configured.\nAborted!\n'

        self.invoke('storage', '--location', 'loc', 'list', 'foo/bar')
        m_get_interface.assert_called_once_with('loc')

    def test_bad_location(self, m_get_interface):
        m_get_interface.side_effect = KeyError
        results = self.invoke('storage', '--location', 'loc', 'list', 'foo/bar', exit_code=1)
        assert results.output == 'The location loc does not exist. ' \
            'Pass --location or change your configuration.\nAborted!\n'


@mock.patch.object(current_app.storage, 'get_interface', autospec=True, spec_set=True)
class TestList(CLIBase):
    cmd_name = 'storage --location loc list'

    def test_list(self, m_get_interface):
        m_list = mock.MagicMock(
            return_value=[LazyDict(last_modified=date(2017, 4, 1), name='foo/bar', size=1024 * 3)]
        )
        m_get_interface.return_value.list = m_list

        results = self.invoke('foo')
        assert results.output == 'Apr 01 2017\t3.0K\tfoo/bar\n'
        m_list.assert_called_once_with('foo')

    def test_list_simple(self, m_get_interface):
        m_list = mock.MagicMock(
            return_value=[LazyDict(last_modified=date(2017, 4, 1), name='foo/bar', size=1024 * 3)]
        )
        m_get_interface.return_value.list = m_list

        results = self.invoke('--simple')
        assert results.output == 'foo/bar\n'
        m_list.assert_called_once_with('/')


@contextlib.contextmanager
def change_dir(path):
    cwd = os.getcwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(cwd)


@mock.patch.object(current_app.storage, 'get_interface', autospec=True, spec_set=True)
class TestGet(CLIBase):
    cmd_name = 'storage --location loc get'

    def test_get_default_dest(self, m_get_interface, tmp_path):
        m_download = mock.MagicMock()
        m_get_interface.return_value.download = m_download

        with change_dir(tmp_path):
            results = self.invoke('foo/bar')

        assert results.output == 'Downloaded foo/bar to bar.\n'

        out_path = os.path.join(tmp_path, 'bar')
        assert os.path.exists(out_path)
        args, _ = m_download.call_args
        assert args[0] == 'foo/bar'
        assert args[1].name == 'bar'

    def test_get_given_dest(self, m_get_interface, tmp_path):
        m_download = mock.MagicMock()
        m_get_interface.return_value.download = m_download

        out_path = os.path.join(tmp_path, 'output')
        results = self.invoke('foo/bar', out_path)
        assert results.output == 'Downloaded foo/bar to {}.\n'.format(out_path)

        args, _ = m_download.call_args
        assert args[0] == 'foo/bar'
        assert args[1].name == out_path

    def test_get_to_stdout(self, m_get_interface):
        m_download = mock.MagicMock()
        m_download.side_effect = lambda path, obj: obj.write(b'test output\n')

        m_get_interface.return_value.download = m_download

        results = self.invoke('foo/bar', '-')
        assert results.output == 'test output\nDownloaded foo/bar to -.\n'

    def test_get_file_not_found(self, m_get_interface):
        m_get_interface.return_value.download.side_effect = \
            FileNotFoundInStorageError('abc', 'def')

        results = self.invoke('foo/bar', 'bar', exit_code=1)
        assert results.output == "Error: Could not open file 'def': Not found in abc.\n"


@mock.patch.object(current_app.storage, 'get_interface', autospec=True, spec_set=True)
class TestPut(CLIBase):
    cmd_name = 'storage --location loc put'

    def test_put(self, m_get_interface, tmp_path):
        m_upload = mock.MagicMock()
        m_get_interface.return_value.upload = m_upload

        source = os.path.join(tmp_path, 'bar.txt')
        with open(source, 'wb'):
            pass

        results = self.invoke(source, 'baz/bar')
        assert results.output == 'Uploaded {} to baz/bar.\n'.format(source)

        args, _ = m_upload.call_args
        assert args[0].name == source
        assert args[1] == 'baz/bar'

    def test_put_from_stdin(self, m_get_interface):
        m_upload = mock.MagicMock()
        stdin_data = io.BytesIO()
        m_upload.side_effect = lambda obj, path: stdin_data.write(obj.read())
        m_get_interface.return_value.upload = m_upload

        results = self.invoke('-', 'baz/bar', input=b'test data')
        assert results.output == 'Uploaded - to baz/bar.\n'

        args, _ = m_upload.call_args
        assert isinstance(args[0], io.BytesIO)
        assert args[1] == 'baz/bar'
        assert stdin_data.getvalue() == b'test data'


@mock.patch.object(current_app.storage, 'get_interface', autospec=True, spec_set=True)
class TestDelete(CLIBase):
    cmd_name = 'storage --location loc delete'

    def test_delete(self, m_get_interface):
        m_delete = mock.MagicMock()
        m_get_interface.return_value.delete = m_delete

        results = self.invoke('foo/bar')
        assert results.output == 'Deleted foo/bar.\n'
        m_delete.assert_called_once_with('foo/bar')

    def test_delete_file_not_found(self, m_get_interface):
        m_get_interface.return_value.delete.side_effect = FileNotFoundInStorageError('abc', 'def')

        results = self.invoke('foo/bar', exit_code=1)
        assert results.output == "Error: Could not open file 'def': Not found in abc.\n"


@mock.patch.object(current_app.storage, 'get_interface', autospec=True, spec_set=True)
class TestLink(CLIBase):
    cmd_name = 'storage --location loc link'

    @freezegun.freeze_time('2020-04-27')
    def test_link_for_default_expiration(self, m_get_interface):
        m_link = mock.MagicMock(return_value='http://example.com/foo/bar')
        m_get_interface.return_value.link_to = m_link

        results = self.invoke('foo/bar')
        assert results.output == 'http://example.com/foo/bar\n'
        m_link.assert_called_once_with(
            path='foo/bar',
            operation=ShareLinkOperation.download,
            expire=arrow.get(2020, 4, 27, 1)
        )

    @freezegun.freeze_time('2020-04-27')
    def test_link_for_expiration_given(self, m_get_interface):
        m_link = mock.MagicMock(return_value='http://example.com/foo/bar')
        m_get_interface.return_value.link_to = m_link

        results = self.invoke('-e', '100', 'foo/bar')
        assert results.output == 'http://example.com/foo/bar\n'
        m_link.assert_called_once_with(
            path='foo/bar',
            operation=ShareLinkOperation.download,
            expire=arrow.get(2020, 4, 27, 0, 1, 40)
        )

    @pytest.mark.parametrize('flags,ops', [
        (['--download'], ShareLinkOperation.download),
        (['--upload'], ShareLinkOperation.download | ShareLinkOperation.upload),
        (['--no-download', '--upload'], ShareLinkOperation.upload),
        (['--no-download', '--delete'], ShareLinkOperation.remove),
        (['--no-download', '--upload', '--delete'],
         ShareLinkOperation.upload | ShareLinkOperation.remove),
        (['--upload', '--delete'],
         ShareLinkOperation.download | ShareLinkOperation.upload | ShareLinkOperation.remove),
    ])
    @freezegun.freeze_time('2020-04-27')
    def test_link_for_operations(self, m_get_interface, flags, ops):
        m_link = mock.MagicMock(return_value='http://example.com/foo/bar')
        m_get_interface.return_value.link_to = m_link

        results = self.invoke(*flags, 'foo/bar')
        assert results.output == 'http://example.com/foo/bar\n'
        m_link.assert_called_once_with(
            path='foo/bar',
            operation=ops,
            expire=arrow.get(2020, 4, 27, 1)
        )

    def test_link_for_error(self, m_get_interface):
        m_get_interface.return_value.link_to.side_effect = ValueError('Some error')

        results = self.invoke('foo/bar', exit_code=1)
        assert results.output == 'Some error\nAborted!\n'


@mock.patch.object(current_app.storage, 'get_interface', autospec=True, spec_set=True)
@mock.patch('keg_storage.utils.reencrypt', autospec=True, spec_set=True)
class TestReencrypt(CLIBase):
    class InputRunner(click.testing.CliRunner):
        def isolation(self, **kwargs):
            kwargs['input'] = 'abc\nxyz'
            return super(self.__class__, self).isolation(**kwargs)

    def test_reencrypt(self, m_reencrypt, m_get_interface):
        m_get_interface.return_value = 'STORAGE'
        app_cls = current_app._get_current_object().__class__
        results = invoke_command(app_cls, 'storage', '--location', 'loc', 'reencrypt', 'foo',
                                 runner=self.InputRunner())

        assert results.output == 'Old Key: \nNew Key: \nRe-encrypted foo\n'
        m_reencrypt.assert_called_once_with('STORAGE', 'foo', b'abc', b'xyz')

    def test_file_not_found(self, m_reencrypt, m_get_interface):
        m_get_interface.return_value = 'STORAGE'
        m_reencrypt.side_effect = FileNotFoundInStorageError('abc', 'def')

        app_cls = current_app._get_current_object().__class__
        results = invoke_command(app_cls, 'storage', '--location', 'loc', 'reencrypt', 'foo',
                                 runner=self.InputRunner(), exit_code=1)

        assert results.output.splitlines() == [
            'Old Key: ',
            'New Key: ',
            "Error: Could not open file 'def': Not found in abc."
        ]



from importlib.util import find_spec

from .base import (
    FileMode,
    FileNotFoundInStorageError,
    ProgressCallback,
    RemoteFile,
    StorageBackend,
    InternalLinksStorageBackend,
    InternalLinkTokenData,
    ShareLinkOperation,
)
from .filesystem import (
    LocalFSStorage,
    LocalFSError,
)

__all__ = [
    "FileMode",
    "FileNotFoundInStorageError",
    "ProgressCallback",
    "RemoteFile",
    "StorageBackend",
    "base",
    "filesystem",
    "LocalFSStorage",
    "LocalFSError",
    "InternalLinksStorageBackend",
    "InternalLinkTokenData",
    "ShareLinkOperation",
]

if find_spec('boto3') is not None:
    from .s3 import S3Storage  # noqa
    __all__.extend(['s3', 'S3Storage'])

if find_spec('paramiko') is not None:
    from .sftp import SFTPStorage  # noqa
    __all__.extend(['sftp', 'SFTPStorage'])

if find_spec('azure') is not None:
    from .azure import AzureStorage  # noqa
    __all__.extend(['azure', 'AzureStorage'])



import io
import os
import pathlib

import arrow
import flask
import flask_webtest
import freezegun
import pytest

from keg_storage import ShareLinkOperation
from keg_storage_ta.views import create_local_storage


class TestViewMixin:
    def setup(self):
        self.client = flask_webtest.TestApp(flask.current_app)

    def test_get_no_token(self, tmp_path: pathlib.Path):
        url = flask.url_for('public.link-view')

        resp = self.client.get(url, headers={'StorageRoot': str(tmp_path)}, status=404)
        assert resp.status_code == 404

        resp = self.client.get(f'{url}?token', headers={'StorageRoot': str(tmp_path)}, status=404)
        assert resp.status_code == 404

    def test_expired_token(self, tmp_path: pathlib.Path):
        storage = create_local_storage(tmp_path)

        now = arrow.utcnow()
        url = storage.link_to(
            path='abc.txt',
            operation=ShareLinkOperation.download,
            expire=now.shift(hours=1)
        )

        with freezegun.freeze_time(now.shift(hours=1, seconds=1).datetime):
            resp = self.client.get(url, headers={'StorageRoot': str(tmp_path)}, status=403)
        assert resp.status_code == 403

    def test_bad_signature(self, tmp_path: pathlib.Path):
        storage = create_local_storage(tmp_path)
        storage.name = 'foo'

        url = storage.link_to(
            path='abc.txt',
            operation=ShareLinkOperation.download,
            expire=arrow.utcnow().shift(hours=1)
        )

        resp = self.client.get(url, headers={'StorageRoot': str(tmp_path)}, status=403)
        assert resp.status_code == 403

    def test_get_operation_not_allowed(self, tmp_path: pathlib.Path):
        storage = create_local_storage(tmp_path)

        url = storage.link_to(
            path='abc.txt',
            operation=ShareLinkOperation.upload,
            expire=arrow.utcnow().shift(hours=1)
        )

        resp = self.client.get(url, headers={'StorageRoot': str(tmp_path)}, status=403)
        assert resp.status_code == 403

    def test_get_success(self, tmp_path: pathlib.Path):
        storage = create_local_storage(tmp_path)
        storage.upload(io.BytesIO(b'foo'), 'abc.txt')
        assert tmp_path.joinpath('abc.txt').exists()

        url = storage.link_to(
            path='abc.txt',
            operation=ShareLinkOperation.download,
            expire=arrow.utcnow().shift(hours=1)
        )

        resp = self.client.get(url, headers={'StorageRoot': str(tmp_path)})
        assert resp.status_code == 200
        assert resp.content_type == 'application/octet-stream'
        assert resp.body == b'foo'

    @pytest.mark.parametrize('method', ['post', 'put'])
    def test_post_put_operation_not_allowed(self, tmp_path: pathlib.Path, method: str):
        storage = create_local_storage(tmp_path)

        url = storage.link_to(
            path='abc.txt',
            operation=ShareLinkOperation.download,
            expire=arrow.utcnow().shift(hours=1)
        )

        req_method = getattr(self.client, method)

        resp = req_method(url, 'foo', headers={'StorageRoot': str(tmp_path)}, status=403)
        assert resp.status_code == 403
        assert not tmp_path.joinpath('abc.txt').exists()

    @pytest.mark.parametrize('method', ['post', 'put'])
    def test_post_put_success(self, tmp_path: pathlib.Path, method):
        storage = create_local_storage(tmp_path)

        assert not tmp_path.joinpath('abc.txt').exists()

        url = storage.link_to(
            path='abc.txt',
            operation=ShareLinkOperation.upload,
            expire=arrow.utcnow().shift(hours=1)
        )

        req_method = getattr(self.client, method)

        resp = req_method(url, 'foo', headers={'StorageRoot': str(tmp_path)})
        assert resp.status_code == 200
        assert resp.body == b'OK'

        with open(tmp_path / 'abc.txt', 'rb') as fp:
            assert fp.read() == b'foo'

    def test_delete_operation_not_allowed(self, tmp_path: pathlib.Path):
        storage = create_local_storage(tmp_path)
        storage.upload(io.BytesIO(b'foo'), 'abc.txt')

        url = storage.link_to(
            path='abc.txt',
            operation=ShareLinkOperation.download,
            expire=arrow.utcnow().shift(hours=1)
        )

        resp = self.client.delete(url, headers={'StorageRoot': str(tmp_path)}, status=403)
        assert resp.status_code == 403
        assert tmp_path.joinpath('abc.txt').exists()

    def test_delete_success(self, tmp_path: pathlib.Path):
        storage = create_local_storage(tmp_path)
        storage.upload(io.BytesIO(b'foo'), 'abc.txt')

        assert tmp_path.joinpath('abc.txt').exists()

        url = storage.link_to(
            path='abc.txt',
            operation=ShareLinkOperation.remove,
            expire=arrow.utcnow().shift(hours=1)
        )

        resp = self.client.delete(url, headers={'StorageRoot': str(tmp_path)})
        assert resp.status_code == 200
        assert resp.body == b'OK'

        assert not tmp_path.joinpath('abc.txt').exists()

    def test_multi_operation(self, tmp_path: pathlib.Path):
        storage = create_local_storage(tmp_path)
        full_path = tmp_path / 'abc.txt'

        url = storage.link_to(
            path='abc.txt',
            operation=(
                ShareLinkOperation.download | ShareLinkOperation.upload | ShareLinkOperation.remove
            ),
            expire=arrow.utcnow().shift(hours=1)
        )

        data = os.urandom(20 * 1024 * 1024)

        resp = self.client.post(url, data, headers={'StorageRoot': str(tmp_path)})
        assert resp.status_code == 200
        with full_path.open('rb') as fp:
            assert fp.read() == data

        resp = self.client.get(url, headers={'StorageRoot': str(tmp_path)})
        assert resp.status_code == 200
        assert resp.body == data

        resp = self.client.delete(url, headers={'StorageRoot': str(tmp_path)})
        assert resp.status_code == 200
        assert not full_path.exists()



from unittest import mock

import pytest

import keg_storage


class TestStorage:
    def test_app_to_init_call_init(self):
        app = mock.MagicMock()
        app.config = {
            "KEG_STORAGE_PROFILES": [(keg_storage.backends.StorageBackend, {"name": "test"})]
        }
        storage = keg_storage.Storage(app)
        assert "test" in storage._interfaces
        assert storage.interface == "test"

        # Test plugin lookup.
        assert isinstance(storage.get_interface(), keg_storage.backends.StorageBackend)
        assert isinstance(storage.get_interface("test"), keg_storage.backends.StorageBackend)

        # Test invalid plugin.
        with pytest.raises(ValueError, match="invalid interface 'foo'"):
            storage.get_interface("foo")

    def test_migration_storage_profiles(self):
        # Old name gets translated to current name.
        app = mock.MagicMock()
        app.config = {
            "STORAGE_PROFILES": [(keg_storage.backends.StorageBackend, {"name": "found"})]
        }

        with pytest.warns(DeprecationWarning, match="STORAGE_PROFILES is deprecated"):
            storage = keg_storage.Storage(app)
        assert "found" in storage._interfaces
        assert storage.interface == "found"

        # If both are there, use the current name.
        app = mock.MagicMock()
        app.config = {
            "STORAGE_PROFILES": [(keg_storage.backends.StorageBackend, {"name": "ignored"})],
            "KEG_STORAGE_PROFILES": [(keg_storage.backends.StorageBackend, {"name": "found"})],
        }

        with pytest.warns(
            DeprecationWarning,
            match="Found both KEG_STORAGE_PROFILES and deprecated STORAGE_PROFILES",
        ):
            storage = keg_storage.Storage(app)
        assert "ignored" not in storage._interfaces
        assert "found" in storage._interfaces
        assert storage.interface == "found"

    def test_no_storage_profiles(self):
        app = mock.MagicMock()
        app.config = {"KEG_STORAGE_PROFILES": []}
        storage = keg_storage.Storage(app)

        with pytest.raises(ValueError, match="no interface was specified"):
            storage.get_interface()
        with pytest.raises(ValueError, match="invalid interface 'foo'"):
            storage.get_interface("foo")



import base64
import os
import typing
import urllib.parse
import warnings
from datetime import datetime
from typing import ClassVar, List, Optional

import arrow
from azure.storage.blob import (
    BlobBlock,
    BlobClient,
    BlobPrefix,
    BlobSasPermissions,
    BlobServiceClient,
    ContainerClient,
    generate_blob_sas,
    generate_container_sas,

)

from keg_storage.backends import base


DEFAULT_CHUNK_SIZE = 5 * 1024 * 1024


class AzureFile(base.RemoteFile):
    """
    Base class for Azure file interface. Since read and write operations are very different and
    integrating the two would introduce a lot of complexity there are distinct subclasses for files
    opened for reading and writing.
    """

    def __init__(self, mode: base.FileMode, blob_client: BlobClient, chunk_size=DEFAULT_CHUNK_SIZE):
        """
        :param mode: file mode
        :param blob_client: blob client instance to use for API calls
        :param chunk_size: read/write buffer size
        """
        super().__init__(mode)
        self.chunk_size = chunk_size
        self.client = blob_client

        # Local buffer to reduce number of requests
        self.buffer = bytearray()


class AzureWriter(AzureFile):
    """
    We are using Azure Block Blobs for all operations. The process for writing them is substantially
    similar to that of S3 with a couple of differences.

    1. We generate the IDs for the blocks
    2. There is no separate call to instantiate the upload. The first call to put_block will create
        the blob.

    """

    max_block_size: ClassVar[int] = 100 * 1024 * 1024

    def __init__(
        self,
        mode: base.FileMode,
        blob_client: BlobClient,
        chunk_size=DEFAULT_CHUNK_SIZE,
    ):
        if chunk_size is not None:
            # chunk_size cannot be larger than max_block_size due to API restrictions
            chunk_size = min(chunk_size, self.max_block_size)
        super().__init__(
            mode=mode,
            blob_client=blob_client,
            chunk_size=chunk_size,
        )
        self.blocks: List[BlobBlock] = []

    def _gen_block_id(self) -> str:
        """
        Generate a unique ID for the block. This is meant to be opaque but it is generated from:
            1. The index of the block as an 64 bit unsigned big endian integer
            2. 40 bytes of random data
        The two parts are concatenated and base64 encoded giving us 64 bytes which is the maximum
        Azure allows.
        """
        index_part = len(self.blocks).to_bytes(8, byteorder='big', signed=False)
        random_part = os.urandom(40)
        return base64.b64encode(index_part + random_part).decode()

    def _flush(self):
        if len(self.buffer) == 0:
            # If there is no buffered data, we don't need to do anything
            return

        # Upload at most chunk_size bytes to a new block
        block_id = self._gen_block_id()
        self.client.stage_block(block_id=block_id, data=bytes(self.buffer[:self.chunk_size]))

        # Store the block_id to later concatenate when we close this file
        self.blocks.append(BlobBlock(block_id=block_id))

        # Cycle the buffer
        self.buffer = self.buffer[self.chunk_size:]

    def _finalize(self):
        self.client.commit_block_list(block_list=self.blocks)
        self.blocks = []

    def write(self, data: bytes) -> None:
        self.buffer.extend(data)
        while len(self.buffer) >= self.chunk_size:
            # Write may be bigger than the chunk size and _flush() only uploads a single chunk so
            # repeated calls may be necessary
            self._flush()

    def close(self):
        self._flush()
        if self.blocks:
            # If we haven't created any blocks, we don't need to finalize
            self._finalize()


class AzureReader(AzureFile):
    """
    The Azure reader uses byte ranged API calls to fill a local buffer to avoid lots of API overhead
    for small read sizes.
    """

    def __init__(
        self,
        mode: base.FileMode,
        blob_client: BlobClient,
        chunk_size=DEFAULT_CHUNK_SIZE,
    ):
        super().__init__(
            mode=mode,
            blob_client=blob_client,
            chunk_size=chunk_size,
        )
        self.stream = self.client.download_blob()
        self.chunks = self.stream.chunks()

    def _read_from_buffer(self, max_size):
        """
        Read up to max_size bytes from the local buffer.
        """
        read_size = min(len(self.buffer), max_size)
        output = self.buffer[:read_size]
        self.buffer = self.buffer[read_size:]
        return output

    def read(self, size: int) -> bytes:
        output_buf = bytes()

        while len(output_buf) < size:
            if len(self.buffer) == 0:
                try:
                    # Load the next chunk into the local buffer
                    next_chunk = next(self.chunks)
                    self.buffer.extend(next_chunk)
                except StopIteration:
                    # All chunks have been consumed
                    break

            read_remainder = size - len(output_buf)
            output_buf += self._read_from_buffer(read_remainder)

        return output_buf


class AzureStorage(base.StorageBackend):
    account_url: Optional[str]
    container_url: Optional[str]
    blob_url: Optional[str]

    def __init__(
        self,
        account: Optional[str] = None,
        key: Optional[str] = None,
        bucket: Optional[str] = None,
        sas_container_url: Optional[str] = None,
        sas_blob_url: Optional[str] = None,
        chunk_size=DEFAULT_CHUNK_SIZE,
        name: str = "azure",
    ):
        super().__init__(name)
        self.chunk_size = chunk_size

        self.account = account
        self.key = key
        self.bucket = bucket

        self.account_url = None
        self.container_url = None
        self.blob_url = None

        if account and key and bucket:
            self.account_url = 'https://{}.blob.core.windows.net'.format(self.account)
        elif sas_container_url:
            self.container_url = sas_container_url
        elif sas_blob_url:
            self.blob_url = sas_blob_url
        else:
            raise ValueError(
                "Must provide a sas_container_url, a sas_blob_url, "
                "or a combination of account, key, and bucket"
            )

    def _create_container_client(self) -> ContainerClient:
        """Create a ContainerClient.

        First see if a ``container_url`` was configured. Otherwise fall back to credentials.
        """

        if self.container_url:
            return ContainerClient.from_container_url(self.container_url)

        service_client = BlobServiceClient(account_url=self.account_url, credential=self.key)
        return service_client.get_container_client(self.bucket)

    def _create_blob_client(self, path: str) -> BlobClient:
        """Create a BlobClient for the given path.

        First see if a ``blob_url`` was configured. Otherwise fall back to ``container_url`` or
        credentials.
        """

        if self.blob_url:
            blob_client = BlobClient.from_blob_url(self.blob_url)
            if blob_client.blob_name != path:
                raise ValueError("Invalid path for the configured SAS blob URL")
            return blob_client

        container_client = self._create_container_client()
        return container_client.get_blob_client(path)

    def _clean_path(self, path: str):
        return path.lstrip('/')

    def list(self, path: str) -> typing.List[base.ListEntry]:
        if self.blob_url:
            raise ValueError("Cannot perform list operation when configured with SAS blob URL")
        client = self._create_container_client()

        if not path.endswith('/'):
            path = path + '/'
        path = self._clean_path(path)
        list_iter = client.walk_blobs(path)

        def construct_entry(blob):
            if isinstance(blob, BlobPrefix):
                return base.ListEntry(name=blob.name, last_modified=None, size=0)
            return base.ListEntry(
                name=blob.name,
                last_modified=arrow.get(blob.last_modified),
                size=blob.size,
            )

        return [construct_entry(blob) for blob in list_iter]

    def open(self, path: str, mode: typing.Union[base.FileMode, str]) -> AzureFile:
        mode = base.FileMode.as_mode(mode)

        path = self._clean_path(path)
        blob_client = self._create_blob_client(path)

        if (mode & base.FileMode.read) and (mode & base.FileMode.write):
            raise NotImplementedError('Read+write mode not supported by the Azure backend')
        elif mode & base.FileMode.write:
            return AzureWriter(mode=mode, blob_client=blob_client, chunk_size=self.chunk_size)
        elif mode & base.FileMode.read:
            return AzureReader(mode=mode, blob_client=blob_client, chunk_size=self.chunk_size)
        else:
            raise ValueError('Unsupported mode. Accepted modes are FileMode.read or FileMode.write')

    def delete(self, path: str):
        path = self._clean_path(path)
        blob_client = self._create_blob_client(path)
        blob_client.delete_blob()

    def create_upload_url(self, path: str, expire: typing.Union[arrow.Arrow, datetime]):
        """
        Create an SAS URL that can be used to upload a blob without any additional authentication.
        This url can be used in following way to authenticate a client and upload to the
        pre-specified path:

            client = BlobClient.from_blob_url(url)
            client.upload_blob(data)
        """
        warnings.warn('create_upload_url is deprecated. Use link_to instead', DeprecationWarning)
        return self.link_to(
            path=path,
            operation=base.ShareLinkOperation.upload,
            expire=expire
        )

    def create_download_url(self, path: str, expire: typing.Union[arrow.Arrow, datetime]):
        """
        Create an SAS URL that can be used to download a blob without any additional authentication.
        This url may be accessed directly to download the blob:

            requests.get(url)
        """
        warnings.warn('create_download_url is deprecated. Use link_to instead', DeprecationWarning)
        return self.link_to(
            path=path,
            operation=base.ShareLinkOperation.download,
            expire=expire
        )

    def link_to(
            self,
            path: str,
            operation: typing.Union[base.ShareLinkOperation, str],
            expire: typing.Union[arrow.Arrow, datetime],
    ) -> str:
        if not self.account_url or not self.key:
            raise ValueError('Cannot create a SAS URL without account credentials')
        path = self._clean_path(path)
        expire = expire.datetime if isinstance(expire, arrow.Arrow) else expire

        operation = base.ShareLinkOperation.as_operation(operation)
        perms = BlobSasPermissions(
            read=operation & base.ShareLinkOperation.download,
            add=False,  # only useful for append blobs that are not supported
            create=operation & base.ShareLinkOperation.upload,
            write=operation & base.ShareLinkOperation.upload,
            delete=operation & base.ShareLinkOperation.remove,
            tag=False
        )

        token = generate_blob_sas(
            account_name=self.account,
            container_name=self.bucket,
            blob_name=path,
            account_key=self.key,
            permission=perms,
            expiry=expire,
        )
        escaped_path = urllib.parse.quote(path, safe="")
        url = urllib.parse.urljoin(self.account_url, '{}/{}'.format(self.bucket, escaped_path))
        return '{}?{}'.format(url, token)

    def create_container_url(self, expire: typing.Union[arrow.Arrow, datetime],
                             ip: typing.Optional[str] = None):
        if not self.account_url:
            raise ValueError('Cannot create a SAS URL without account credentials')
        expire = expire.datetime if isinstance(expire, arrow.Arrow) else expire
        token = generate_container_sas(
            account_name=self.account,
            container_name=self.bucket,
            account_key=self.key,
            permission='rwdl',
            expiry=expire,
            ip=ip
        )
        url = urllib.parse.urljoin(self.account_url, self.bucket)
        return '{}?{}'.format(url, token)



import enum
import typing
from datetime import datetime

import arrow
import itsdangerous

from keg_storage.utils import expire_time_to_seconds

ProgressCallback = typing.Callable[[int], None]


class ListEntry(typing.NamedTuple):
    name: str
    last_modified: arrow.Arrow
    size: int


class FileMode(enum.Flag):
    read = enum.auto()
    write = enum.auto()

    def __str__(self):
        s = 'r' if self & FileMode.read else ''
        s += 'w' if self & FileMode.write else ''
        return f'{s}b'

    @classmethod
    def as_mode(cls, obj: typing.Union[str, 'FileMode']) -> 'FileMode':
        if isinstance(obj, cls):
            return obj
        if not isinstance(obj, str):
            raise ValueError('as_mode() accepts only FileMode or str arguments')

        mode = cls(0)
        if 'r' in obj:
            mode |= cls.read
        if 'w' in obj:
            mode |= cls.write
        return mode


class ShareLinkOperation(enum.Flag):
    download = enum.auto()
    upload = enum.auto()
    remove = enum.auto()

    @classmethod
    def as_operation(cls, obj: typing.Union[str, 'ShareLinkOperation']) -> 'ShareLinkOperation':
        if isinstance(obj, cls):
            return obj
        if not isinstance(obj, str):
            raise ValueError(f'as_operation() accepts only {cls.__name__} or str arguments')

        op = cls(0)
        if 'd' in obj:
            op |= cls.download
        if 'u' in obj:
            op |= cls.upload
        if 'r' in obj:
            op |= cls.remove
        return op

    def __str__(self):
        return ''.join([
            'd' if self & ShareLinkOperation.download else '',
            'u' if self & ShareLinkOperation.upload else '',
            'r' if self & ShareLinkOperation.remove else '',
        ])


class RemoteFile:
    """
    This is a base class for objects returned by a backend's `open()` method. This is a file-like
    object that provides read/write operations to the remote file. When creating a new backend, you
    should subclass this and implement `read()` and `write()` methods at minimum.

    After construction, a RemoteFile is presumed to be in an "open" state and should accept calls
    to any of its methods. Any cleanup should be done in the `close()` method.
    """

    # This is the default chunk size to use when iterating over this object
    iter_chunk_size = 5 * 1024 * 1024

    def __init__(self, mode: FileMode):
        """
        Override this constructor to accept any additional arguments needed by the backend and to
        perform any initialization required to get the file into an "open" state.
        """
        self.mode = mode

    def read(self, size: int) -> bytes:
        """
        Read and return up to `size` bytes from the remote file. If the end of the file is reached
        this should return an empty bytes string.
        """
        raise NotImplementedError

    def write(self, data: bytes) -> None:
        """
        Write the data buffer to the remote file.
        """
        raise NotImplementedError

    def close(self):
        """
        Cleanup and deallocate any held resources. This method may be called multiple times on a
        single instance. If the file was already closed, this method should do nothing.
        """
        pass

    def iter_chunks(self, chunk_size: int = None):
        """
        Iterate over the file in blocks of `chunk_size`.
        """
        chunk_size = chunk_size or self.iter_chunk_size

        while True:
            chunk = self.read(chunk_size)
            if chunk == b'':
                break
            yield chunk

    def __iter__(self):
        return self.iter_chunks()

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class StorageBackend:
    def __init__(self, name: str):
        self.name = name

    def list(self, path: str) -> typing.List[ListEntry]:
        """
        Returns a list of `ListEntry`s representing files available under the directory or prefix
        given in `path`.
        """
        raise NotImplementedError()

    def open(self, path: str, mode: typing.Union[FileMode, str]) -> RemoteFile:
        """
        Returns a instance of RemoteFile for the given `path` that can be used for
        reading and/or writing depending on the `mode` given.
        """
        raise NotImplementedError()

    def delete(self, path: str):
        """
        Delete the remote file specified by `path`.
        """
        raise NotImplementedError()

    def copy(self, path: str, new_path: str):
        """
        Copy the remote file specified by `path` to `new_path`.
        """
        raise NotImplementedError()

    def link_to(
            self,
            path: str,
            operation: typing.Union[ShareLinkOperation, str],
            expire: typing.Union[arrow.Arrow, datetime]
    ) -> str:
        """
        Returns a URL allowing direct the specified operations to be performed on the given path
        """
        raise NotImplementedError()

    def get(self, path: str, dest: str) -> None:
        """
        Copies a remote file at `path` to the `dest` path given on the local filesystem.
        """
        with open(dest, str(FileMode.write)) as outfile:
            self.download(path, outfile)

    def download(
        self,
        path: str,
        file_obj: typing.IO,
        *,
        progress_callback: typing.Optional[ProgressCallback] = None
    ):
        """
        Copies a remote file at `path` to a file-like object `file_obj`.

        If desired, a progress callback can be supplied. The function should accept an int
        parameter, which will be the number of bytes downloaded so far.
        """
        bytes_read = 0

        with self.open(path, FileMode.read) as infile:
            for chunk in infile.iter_chunks():
                file_obj.write(chunk)
                bytes_read += len(chunk)
                if progress_callback:
                    progress_callback(bytes_read)

    def put(self, path: str, dest: str) -> None:
        """
        Copies a local file at `path` to a remote file at `dest`.
        """
        with open(path, str(FileMode.read)) as infile:
            self.upload(infile, dest)

    def upload(
        self,
        file_obj: typing.IO,
        path: str,
        *,
        progress_callback: typing.Optional[ProgressCallback] = None
    ):
        """
        Copies the contents of a file-like object `file_obj` to a remote file at `path`

        If desired, a progress callback can be supplied. The function should accept an int
        parameter, which will be the number of bytes uploaded so far.
        """
        bytes_written = 0
        buffer_size = 5 * 1024 * 1024

        with self.open(path, FileMode.write) as outfile:
            buf = file_obj.read(buffer_size)
            while buf:
                outfile.write(buf)
                bytes_written += len(buf)
                if progress_callback:
                    progress_callback(bytes_written)
                buf = file_obj.read(buffer_size)

    def __str__(self):
        return self.__class__.__name__


class InternalLinkTokenData(typing.NamedTuple):
    path: str
    operations: ShareLinkOperation

    def serialize(self) -> typing.Mapping[str, str]:
        return {
            'key': self.path,
            'op': str(self.operations)
        }

    @classmethod
    def deserialize(cls, data: typing.Mapping[str, str]) -> 'InternalLinkTokenData':
        return cls(
            path=data['key'],
            operations=ShareLinkOperation.as_operation(data['op'])
        )

    @property
    def allow_upload(self) -> bool:
        return ShareLinkOperation.upload in self.operations

    @property
    def allow_download(self) -> bool:
        return ShareLinkOperation.download in self.operations

    @property
    def allow_remove(self) -> bool:
        return ShareLinkOperation.remove in self.operations


class InternalLinksStorageBackend(StorageBackend):
    """
    Base class for storage backends that do not have their own direct method of creating
    download/upload/deletion URLs. To use the link_to feature for such backends, the app must
    provide it's own endpoint to handle the requests.
    See plugin.LinkViewMixin for a base implementation of such an endpoint.
    """

    def __init__(
            self,
            *,
            linked_endpoint: typing.Optional[str],
            secret_key: typing.Optional[bytes],
            name: str,
    ):
        """
        :param name: Backend name
        :param linked_endpoint: The endpoint that the app uses to handle requests to the link_to URL
        :param secret_key: A secret value used to sign the token in the URL
        """
        super().__init__(name=name)
        self.linked_endpoint = linked_endpoint
        self.secret_key = secret_key

    def create_link_token(
            self,
            *,
            path: str,
            operation: typing.Union[ShareLinkOperation, str],
            expire: typing.Union[arrow.Arrow, datetime]
    ):
        """
        Create a signed JWT authorizing the user to perform the specified operations
        """
        if self.secret_key is None:
            raise ValueError('Backend must be configured with secret_key to use this feature')

        signer = itsdangerous.TimedJSONWebSignatureSerializer(
            secret_key=self.secret_key,
            expires_in=int(expire_time_to_seconds(expire)),
            salt=self.name,
        )
        token_data = InternalLinkTokenData(
            path=path,
            operations=operation
        )
        return signer.dumps(token_data.serialize())

    def deserialize_link_token(self, token: str) -> InternalLinkTokenData:
        """
        Verify a JWT and extract the path and allowed operations
        """
        if self.secret_key is None:
            raise ValueError('Backend must be configured with secret_key to use this feature')

        signer = itsdangerous.TimedJSONWebSignatureSerializer(
            secret_key=self.secret_key,
            salt=self.name,
        )
        data = signer.loads(token)

        return InternalLinkTokenData.deserialize(data)

    def link_to(
            self,
            path: str,
            operation: typing.Union[ShareLinkOperation, str],
            expire: typing.Union[arrow.Arrow, datetime]
    ) -> str:
        """
        Create a URL pointing to the given `linked_endpoint` containing a JWT authorizing the user
        user to perform the given operations.

        This is currently only implemented for flask based apps but you may override this method in
        your own subclass to support other frameworks.

        To use this method you must provide `secret_key` and `linked_endpoint` to the constructor.
        """
        try:
            import flask
        except ImportError:
            raise NotImplementedError(
                'This feature is currently only implemented for flask apps. '
                'You may override the link_to method to generate links for your framework.'
            )

        if self.secret_key is None or self.linked_endpoint is None:
            raise ValueError(
                'Backend must be configured with linked_endpoint and secret_key to use this feature'
            )

        token = self.create_link_token(
            path=path,
            operation=operation,
            expire=expire
        )
        return flask.url_for(self.linked_endpoint, token=token.decode(), _external=True)


class FileNotFoundInStorageError(Exception):
    def __init__(self, storage_type, filename):
        self.storage_type = storage_type
        self.filename = filename

    def __str__(self):
        return "File {} not found in {}.".format(self.filename, str(self.storage_type))
